/// Functions for performing cryptographic operations on the main data structures.

use super::*;
use openssl::{
    symm::{Cipher, encrypt as openssl_encrypt, decrypt as openssl_decrypt}
};
use serde_block_tree::{to_vec, from_vec};
use serde::de::{DeserializeOwned};

/// Data that may or may not be encrypted.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Cryptotext<T: Serialize> {
    /// The inner value of `T` is in plaintext.
    Plain(T),
    /// The inner value of `T` is in ciphertext.
    Cipher(Vec<u8>),
}

pub trait Hydrate<const N: usize> {
    fn hydrate(template: Self, data: [u8; N]) -> Self;
}

/// Errors that can occur during cryptographic operations.
#[derive(Debug)]
pub enum Error {
    NoReadCap,
    Encryption(String),
    Decryption(String),
}

pub type Result<T> = std::result::Result<T, Error>;

/// A cryptographic hash.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
pub enum Hash {
    Sha2_256([u8; 32]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; 64]),
}

/// A cryptographic signature.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Signature {
    #[serde(with = "BigArray")]
    Ed25519([u8; 64]),
}

/// A cryptographic key.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub enum Key {
    // For keys that contain IVs, the key always comes first. The function `key_len` can be used
    // to get the length of the key part of a variant's array.

    /// A key for the AES 256 cipher in Cipher Block Chaining mode. Note that this includes the
    /// initialization vector, so that a value of this variant contains all the information needed
    /// to fully initialize a cipher context.
    #[serde(with = "BigArray")]
    Aes256Cbc([u8; 32 + 16]),
    /// A key for the AES 256 cipher in counter mode.
    Aes256Ctr([u8; 32]),
}

impl Key {
    /// Returns the array in this `Key` as a slice.
    pub fn as_slice(&self) -> &[u8] {
        match self {
            Key::Aes256Cbc(array) => array,
            Key::Aes256Ctr(array) => array,
        }
    }

    /// Returns the array in this `Key` as a mutable slice.
    pub fn as_mut_slice(&mut self) -> &mut [u8] {
        match self {
            Key::Aes256Cbc(array) => array,
            Key::Aes256Ctr(array) => array,
        }
    }

    /// Returns the length of the key part of the array in this `Key`. Note that the bytes of the
    /// key come before the bytes of the IV.
    const fn key_len(&self) -> usize {
        let (array_len, key_len) = match self {
            Key::Aes256Cbc(array) => (array.len(), 32),
            Key::Aes256Ctr(array) => (array.len(), 32)
        };
        debug_assert!(key_len <= array_len);
        key_len
    }

    /// Returns the key part of this `Key`.
    pub fn key_slice(&self) -> &[u8] {
        let key_len = self.key_len();
        &self.as_slice()[..key_len]
    }

    /// Returns the IV part of this`Key`.
    pub fn iv_slice(&self) -> Option<&[u8]> {
        let key_len = self.key_len();
        let slice = self.as_slice();
        if key_len == slice.len() {
            None
        }
        else {
            Some(&slice[key_len..])
        }
    }

    /// Packs the given key and IV into the given output buffer.
    pub fn pack(output: &mut [u8], key: &[u8], iv: &[u8]) {
        debug_assert_eq!(output.len(), key.len() + iv.len());
        let key_len = key.len();
        output[..key_len].copy_from_slice(key);
        output[key_len..].copy_from_slice(iv);
    }

    /// Returns the cipher that this key is for.
    fn cipher(&self) -> Cipher {
        match self {
            Key::Aes256Cbc(_) => Cipher::aes_256_cbc(),
            Key::Aes256Ctr(_) => Cipher::aes_256_ctr(),
        }
    }
}

impl AsRef<[u8]> for Key {
    fn as_ref(&self) -> &[u8] {
        self.as_slice()
    }
}

impl AsMut<[u8]> for Key {
    fn as_mut(&mut self) -> &mut [u8] {
        self.as_mut_slice()
    }
}

#[allow(dead_code)]
pub(crate) fn encrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Cipher(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Plain(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = encrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Cipher(new_body);
    block.read_caps.insert(principal_owned, encrypt(&block_key, key)?);
    Ok(VersionedBlock::V0(block))
}

#[allow(dead_code)]
pub(crate) fn decrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Plain(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Cipher(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = decrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Plain(new_body);
    block.read_caps.insert(principal_owned, Cryptotext::Plain(block_key));
    Ok(VersionedBlock::V0(block))
}

fn encrypt<T: Serialize>(value: &T, key: &Key) -> Result<Cryptotext<T>> {
    let data = to_vec(value).map_err(|err| Error::Encryption(err.to_string()))?;
    let vec = encrypt_slice(data.as_slice(), key);
    Ok(Cryptotext::Cipher(vec?))
}

fn decrypt<T: Serialize + DeserializeOwned>(cryptotext: Cryptotext<T>, key: &Key) -> Result<T> {
    let data = match cryptotext {
        Cryptotext::Plain(value) => return Ok(value),
        Cryptotext::Cipher(data) => data
    };
    let vec = decrypt_slice(data.as_ref(), key);
    let value = from_vec(&vec?).map_err(|err| Error::Decryption(err.to_string()));
    value
}

fn encrypt_slice(plaintext: &[u8], key: &Key) -> Result<Vec<u8>> {
    let cipher = key.cipher();
    let key_slice = key.key_slice();
    let iv_slice = key.iv_slice();
    openssl_encrypt(cipher, key_slice, iv_slice, plaintext)
        .map_err(|err| Error::Encryption(err.to_string()))
}

fn decrypt_slice(ciphertext: &[u8], key: &Key) -> Result<Vec<u8>> {
    let cipher = key.cipher();
    let key_slice = key.key_slice();
    let iv_slice = key.iv_slice();
    openssl_decrypt(cipher, key_slice, iv_slice, ciphertext)
        .map_err(|err| Error::Decryption(err.to_string()))
}

#[cfg(test)]
mod tests {
    use super::*;

    const KEY: [u8; 32] = [
        0xB2, 0xB3, 0xDA, 0x5A, 0x1A, 0xF6, 0xB3, 0x78, 0x30, 0xAB, 0x1D, 0x33, 0x33, 0xE7, 0xE3,
        0x5B, 0xBB, 0xF9, 0xFE, 0xD0, 0xC1, 0xF7, 0x90, 0x34, 0x69, 0xB7, 0xE7, 0xC6, 0x1C, 0x46,
        0x85, 0x48,
    ];

    const IV: [u8; 16] = [
        0x60, 0xE0, 0xBE, 0x45, 0xA9, 0xAB, 0x5D, 0x99, 0x3B, 0x3A, 0x1E, 0x54, 0x18, 0xE0, 0x46,
        0xDE,
    ];

    #[test]
    fn key_aes256cbc_key_part_is_correct() {
        let mut array = [0u8; KEY.len() + IV.len()];
        Key::pack(&mut array, &KEY, &IV);
        let key = Key::Aes256Cbc(array);
        let key_part = key.key_slice();
        assert_eq!(KEY, key_part);
    }

    #[test]
    fn key_aes256cbc_iv_part_is_correct() {
        let mut array = [0u8; KEY.len() + IV.len()];
        Key::pack(&mut array, &KEY, &IV);
        let key = Key::Aes256Cbc(array);
        let iv_part = key.iv_slice().unwrap();
        assert_eq!(IV, iv_part);
    }

    /// Tests that validate the dependencies of this module.
    mod dependency_tests {
        /// This test validates that data decrypted with the `Crypter` API matches data that was
        /// previously encrypted using it.
        #[test]
        fn crypter() {
            use super::*;
            let expected = b"We attack at the crack of noon!";
            let cipher = Cipher::aes_256_cbc();

            let ciphertext = openssl_encrypt(cipher, &KEY, Some(&IV), expected).unwrap();
            let actual = openssl_decrypt(cipher, &KEY, Some(&IV), ciphertext.as_slice()).unwrap();

            assert_eq!(expected, actual.as_slice());
        }
    }
}
